function [LSet, dcSet , scSet, vCSet, hCSet] = CoGrabCut(imSet, initMapSet, shapePriorMapSet)
%
% Performs segmentation of a set of images into foreground/background
% regions
%
% Usage:
%   L = GrabCut(im, initMap)
%
% Inputs:
%   im - color image to be segmented
%   initMap - initial labeling of which we are _certain_ of:
%             1  - FG
%             -1 - BG
%             0  - uncertain
%
% Output:
%   L - a segmentation (i.e., {0,1} labeling) of the give image.
%
% This implementation follows the paper
% Rother C., Kolmogorov V. and Blake A. "GrabCut"-Interactive
% Foreground Extraction using iterated Graph Cuts. SIGGRAPH 2004.
%

% constants:
Gamma = 50;
Beta = 3;
MaxItr = 100;
K = 5; % number of components in each mixture model
INF = 1000;

labimSet = imSet;

vCSet = {};
hCSet = {};
scSet = {};
certain_fg_set = {};
certain_bg_set = {};
oLSet = {};
for i = 1:numel(labimSet)
    labim = labimSet{i};
    [hC vC] = SmoothnessTerm(labim);
    hCSet{i} = hC;
    vCSet{i} = vC;
    scSet{i} = Gamma.*[0 1; 1 0];
    
    initMap = initMapSet{i};
    certain_fg = initMap == 1;
    certain_bg = initMap == -1;

    oL = initMap;

    % how to label all the uncertain pixels?
    if (sum(sum(abs(certain_fg))) == 0)
        oL(initMap==0) = 1;
    else
        oL(initMap==0) = -1;
    end
    
    certain_fg_set{i} = certain_fg;
    certain_bg_set{i} = certain_bg;
    oLSet{i} = oL;
end



done_suc = false;

for itr=1:MaxItr,
    itr

    logpFGSet = CoLocalColorModel(labimSet, K, oLSet, 1); 
    
    LSet = {};
    dcSet = {};
    for i = 1:numel(labimSet)
%         if sum(oLSet{i}(:) == 1) <= K || sum(oLSet{i}(:) == -1) <= K
%             LSet{i} = oLSet{i};
%             continue;
%         end
        % GMM for foreground and background - global model
        
        logpBGSet = CoLocalColorModel(labimSet(i), K, oLSet(i), -1); 
    
        logpFG = logpFGSet{i};
        logpBG = logpBGSet{1};
        certain_bg = certain_bg_set{i};
        
        shapePriorMap = shapePriorMapSet{i};
        logpFG = logpFG - Beta*log(shapePriorMap);
        logpBG = logpBG - Beta*log(1-shapePriorMap);
        
        logpFG(isinf(logpFG)) = INF;
        logpBG(isinf(logpBG)) = INF;
        
        % force labeling of certain labeling
    %   logpBG(certain_fg) = INF;
    %   logpFG(certain_bg) = INF;

        dc = cat(3, logpBG, logpFG);
        dcSet{i} = dc;
        
        sc = scSet{i};
        hC = hCSet{i};
        vC = vCSet{i};
        oL = oLSet{i};

        gch = GraphCut('open', dc , sc, vC, hC);  
        gch = GraphCut('set', gch, int32(oL==1)); % initial guess - previous result
        [gch L] = GraphCut('expand', gch);
        L = (2*L-1); % convert {0,1} to {-1,1} labeling
        gch = GraphCut('close', gch);
        
        LSet{i} = L;
    end

    % stop if converged
    change = 0; pixelSum = 0;
    for i = 1:numel(labimSet);
        change = change + sum(oLSet{i}(:)~=LSet{i}(:));
        pixelSum = pixelSum + numel(LSet{i});
    end
    if change < .001*pixelSum
        done_suc = true;
        break;
    end
    oLSet = LSet;
end
if ~done_suc
    warning('GrabCut:GrabCut','Failed to converge after %d iterations', itr);
end

